@@ -12,6 +12,8 @@ from sklearn import svm
 import sklearn.feature_selection
 from functools import reduce
 
+from os import path
+
 
 class RangeTolerance:
     def __init__(self, value: float, tolerance: float):
@@ -181,9 +183,9 @@ class ClassificationEvolutionaryPattern(IClassificationPlugin):
     def __init__(self):
         super(ClassificationEvolutionaryPattern, self).__init__()
         self.number_of_generations = 100
-        self.population_size = 1000
-        self.mutation_rate = 0.2
-        self.crossover_rate = 0.1
+        self.population_size = 10000
+        self.mutation_rate = 0.1
+        self.crossover_rate = 0.05
         self.number_selected_features = 30  # the number of (best in population) features to select for training
         
         # TODO implement elitism
@@ -338,6 +340,27 @@ class ClassificationEvolutionaryPattern(IClassificationPlugin):
             generation, scores.mean(), scores.std() * 2, avg_score, std_score, number_unique_patterns,
             np.mean(np.array([p.length() for p in population]))))
 
+            file_path = 'output/evolution.txt'
+            if path.exists(file_path):
+                file = open(file_path, "a+")
+            else:
+                file = open(file_path, 'w+')
+                file.write(
+                    "generation scores_mean scores_std*2 avg_score std_score number_unique_patterns mean_length_population")
+            to_save_ = [
+                generation
+                , scores.mean()
+                , scores.std()*2
+                , avg_score
+                , std_score
+                , number_unique_patterns
+                , np.mean(np.array([p.length() for p in population]))
+            ]
+            file.write('\n' + ' '.join(map(str, to_save_)))
+            file.close()
+
+            # TODO save in textfile
+
             population = self.reproduce(population)
 
     def predict(self, data: DataModel) -> int:
