@@ -187,4 +187,6 @@ class EventdetectionIHMM(IEventdetectionPlugin):
         state_sequence, fixation_index = self.predict_state_sequence(velocities)
 
         self.process_state_sequence(state_sequence, fixation_index, data)
-        data.test_events_consistency()
+        #print(data.events)
+        #data.test_events_consistency()
+        #np.savetxt("ihmm.txt", data.events)
